import inspect
import sys
from typing import Any, Type, TypeVar

import numpy as np
import numpy.typing as npt

from overcooked_simulator.counters import (
    Counter,
    CookingCounter,
    Dispenser,
    ServingWindow,
    CuttingBoard,
    PlateDispenser,
    Sink,
    PlateConfig,
    SinkAddon,
)
from overcooked_simulator.game_items import ItemInfo, ItemType, CookingEquipment, Plate
from overcooked_simulator.order import OrderAndScoreManager
from overcooked_simulator.utils import get_closest

T = TypeVar("T")


def convert_words_to_chars(layout_chars_config: dict[str, str]) -> dict[str, str]:
    """Converts words in a given layout chars configuration dictionary to their corresponding characters.

    Args:
        layout_chars_config (dict[str, str]): A dictionary containing layout character configurations, where the keys are words
            representing layout characters and the values are the corresponding character representations.

    Returns:
        dict[str, str]: A dictionary where the keys are the layout characters and the values are their corresponding words.
    """
    word_refs = {
        "underscore": "_",
        "hash": "#",
        "space": " ",
        "dot": ".",
        "comma": ",",
        "semicolon": ";",
        "colon": ":",
        "plus": "+",
        "minus": "-",
        "exclamation": "!",
        "question": "?",
        "dquote": '"',
        "squote": "'",
        "star": "*",
        "dollar": "$",
        "euro": "€",
        "ampersand": "&",
        "slash": "/",
        "oparentheses": "(",
        "cparentheses": ")",
        "equal": "=",
        "right": ">",
        "left": "<",
        "pipe": "|",
        "at": "@",
        "top": "^",
        "tilde": "~",
    }
    return {word_refs.get(c, c): name for c, name in layout_chars_config.items()}


class CounterFactory:
    """The `CounterFactory` class is responsible for creating counter objects based on the layout configuration and
    item information provided. It also provides methods for mapping and filtering the item information.
    """

    additional_counter_names = {"Counter"}

    def __init__(
        self,
        layout_chars_config: dict[str, str],
        item_info: dict[str, ItemInfo],
        serving_window_additional_kwargs: dict[str, Any],
        plate_config: PlateConfig,
        order_and_score: OrderAndScoreManager,
    ) -> None:
        """Constructor for the `CounterFactory` class. Set up the attributes necessary to instantiate the counters.

        Args:
            layout_chars_config (dict[str, str]): A dictionary mapping layout characters to their corresponding names.
            item_info (dict[str, ItemInfo]): A dictionary containing information about different items.
            serving_window_additional_kwargs (dict[str, Any]): Additional keyword arguments for serving window configuration.
            plate_config (PlateConfig): The configuration for plate usage.

        Returns:
            None

        Initializes the object with the provided parameters. It performs the following tasks:
        - Converts the layout character configuration from words to characters.
        - Sets the item information dictionary.
        - Sets the additional keyword arguments for serving window configuration.
        - Sets the plate configuration.

        It also sets the following attributes:
        - `no_counter_chars`: A set of characters that represent counters for agents or free spaces.
        - `counter_classes`: A dictionary of counter classes imported from the 'overcooked_simulator.counters' module.
        - `cooking_counter_equipments`: A dictionary mapping cooking counters to the list of equipment items associated with them.
        """
        self.layout_chars_config = convert_words_to_chars(layout_chars_config)
        self.item_info = item_info
        self.serving_window_additional_kwargs = serving_window_additional_kwargs
        self.plate_config = plate_config
        self.order_and_score = order_and_score

        self.no_counter_chars = set(
            c
            for c, name in self.layout_chars_config.items()
            if name in ["Agent", "Free"]
        )

        self.counter_classes = dict(
            inspect.getmembers(
                sys.modules["overcooked_simulator.counters"], inspect.isclass
            )
        )

        self.cooking_counter_equipments = {
            cooking_counter: [
                equipment
                for equipment, e_info in self.item_info.items()
                if e_info.equipment and e_info.equipment.name == cooking_counter
            ]
            for cooking_counter, info in self.item_info.items()
            if info.type == ItemType.Equipment and info.equipment is None
        }

    def get_counter_object(self, c: str, pos: npt.NDArray[float]) -> Counter:
        """Create and returns a counter object based on the provided character and position."""

        assert self.can_map(c), f"Can't map counter char {c}"
        counter_class = None
        if self.layout_chars_config[c] in self.item_info:
            item_info = self.item_info[self.layout_chars_config[c]]
            if item_info.type == ItemType.Equipment and item_info.equipment:
                if item_info.equipment.name not in self.counter_classes:
                    return CookingCounter(
                        name=item_info.equipment.name,
                        cooking_counter_equipments=self.cooking_counter_equipments,
                        pos=pos,
                        occupied_by=CookingEquipment(
                            name=item_info.name,
                            item_info=item_info,
                            transitions=self.filter_item_info(
                                by_equipment_name=item_info.name
                            ),
                        ),
                    )
            elif item_info.type == ItemType.Ingredient:
                return Dispenser(pos=pos, dispensing=item_info)

        if counter_class is None:
            counter_class = self.counter_classes[self.layout_chars_config[c]]
        kwargs = {
            "pos": pos,
        }
        if issubclass(counter_class, (CuttingBoard, Sink)):
            kwargs["transitions"] = self.filter_item_info(
                by_equipment_name=counter_class.__name__
            )
        elif issubclass(counter_class, PlateDispenser):
            kwargs.update(
                {
                    "plate_transitions": self.filter_item_info(
                        by_item_type=ItemType.Meal
                    ),
                    "plate_config": self.plate_config,
                    "dispensing": self.item_info[Plate.__name__],
                }
            )
        elif issubclass(counter_class, ServingWindow):
            kwargs.update(self.serving_window_additional_kwargs)
            kwargs[
                "order_and_score"
            ] = self.order_and_score  # individual because for the later trash scorer
        return counter_class(**kwargs)

    def can_map(self, char) -> bool:
        """Check if the provided character can be mapped to a counter object."""
        return char in self.layout_chars_config and (
            not self.is_counter(char)
            or self.layout_chars_config[char] in self.item_info
            or self.layout_chars_config[char] in self.counter_classes
        )

    def is_counter(self, c: str) -> bool:
        """Checks if the provided character represents a counter."""
        return c in self.layout_chars_config and c not in self.no_counter_chars

    def map_not_counter(self, c: str) -> str:
        """Maps the provided character to a non-counter word based on the layout configuration."""
        assert self.can_map(c) and not self.is_counter(
            c
        ), "Cannot map char {c} as a 'not counter'"
        return self.layout_chars_config[c]

    def filter_item_info(
        self,
        by_item_type: ItemType = None,
        by_equipment_name: str = None,
    ) -> dict[str, ItemInfo]:
        """Filter the item info dict by item type or equipment name"""
        if by_item_type is not None:
            return {
                name: info
                for name, info in self.item_info.items()
                if info.type == by_item_type
            }
        if by_equipment_name is not None:
            return {
                name: info
                for name, info in self.item_info.items()
                if info.equipment is not None
                and info.equipment.name == by_equipment_name
            }
        return self.item_info

    def post_counter_setup(self, counters: list[Counter]):
        """Initialize the counters in the environment.

        Connect the `ServingWindow`(s) with the `PlateDispenser`.
        Find and connect the `SinkAddon`s with the `Sink`s

        Args:
            counters: list of counters to perform the post setup on.
        """
        plate_dispenser = self.get_counter_of_type(PlateDispenser, counters)
        assert len(plate_dispenser) > 0, "No Plate Dispenser in the environment"

        sink_addons = self.get_counter_of_type(SinkAddon, counters)

        for counter in counters:
            match counter:
                case ServingWindow():
                    counter: ServingWindow  # Pycharm type checker does now work for match statements?
                    counter.add_plate_dispenser(plate_dispenser[0])
                case Sink(pos=pos):
                    counter: Sink  # Pycharm type checker does now work for match statements?
                    assert len(sink_addons) > 0, "No SinkAddon but normal Sink"
                    closest_addon = get_closest(pos, sink_addons)
                    assert 1 - (1 * 0.05) <= np.linalg.norm(
                        closest_addon.pos - pos
                    ), f"No SinkAddon connected to Sink at pos {pos}"
                    counter.set_addon(closest_addon)

    @staticmethod
    def get_counter_of_type(counter_type: Type[T], counters: list[Counter]) -> list[T]:
        """Filter all counters in the environment for a counter type."""
        return list(filter(lambda counter: isinstance(counter, counter_type), counters))
