import argparse
import dataclasses
import json
import logging
import sys
from enum import Enum

import numpy as np
import pygame
import pygame_gui
import requests
import yaml
from websockets.sync.client import connect

from overcooked_simulator import ROOT_DIR
from overcooked_simulator.game_server import CreateEnvironmentConfig
from overcooked_simulator.gui_2d_vis.drawing import Visualizer
from overcooked_simulator.gui_2d_vis.game_colors import colors
from overcooked_simulator.overcooked_environment import (
    Action,
    ActionType,
    InterActionData,
)
from overcooked_simulator.utils import (
    custom_asdict_factory,
    setup_logging,
    url_and_port_arguments,
    disable_websocket_logging_arguments,
)


class MenuStates(Enum):
    Start = "Start"
    Game = "Game"
    End = "End"


MANAGER_ID = "1233245425"


log = logging.getLogger(__name__)


class PlayerKeySet:
    """Set of keyboard keys for controlling a player.
    First four keys are for movement. Order: Down, Up, Left, Right.    5th key is for interacting with counters.    6th key ist for picking up things or dropping them.
    """

    def __init__(self, player_name: str | int, keys: list[pygame.key]):
        """Creates a player key set which contains information about which keyboard keys control the player.

        Movement keys in the following order: Down, Up, Left, Right

        Args:
            player_name: The name of the player to control.
            keys: The keys which control this player in the following order: Down, Up, Left, Right, Interact, Pickup.
        """
        self.name = player_name
        self.player_keys = keys
        self.move_vectors = [[-1, 0], [1, 0], [0, -1], [0, 1]]
        self.key_to_movement = {
            key: vec for (key, vec) in zip(self.player_keys[:-2], self.move_vectors)
        }
        self.interact_key = self.player_keys[-2]
        self.pickup_key = self.player_keys[-1]


class PyGameGUI:
    """Visualisation of the overcooked environment and reading keyboard inputs using pygame."""

    def __init__(
        self,
        player_names: list[str | int],
        player_keys: list[pygame.key],
        url: str,
        port: str,
    ):
        self.game_screen = None
        self.FPS = 60
        self.running = True

        self.player_names = player_names
        self.player_keys = player_keys

        self.player_key_sets: list[PlayerKeySet] = [
            PlayerKeySet(player_name, keys)
            for player_name, keys in zip(
                self.player_names, self.player_keys[: len(self.player_names)]
            )
        ]

        self.websocket_url = f"ws://{url}:{port}/ws/player/"
        self.websockets = {}

        self.request_url = f"http://{url}:{port}"

        # TODO cache loaded images?
        with open(ROOT_DIR / "gui_2d_vis" / "visualization.yaml", "r") as file:
            self.visualization_config = yaml.safe_load(file)

        self.screen_margin = self.visualization_config["GameWindow"]["screen_margin"]
        self.min_width = self.visualization_config["GameWindow"]["min_width"]
        self.min_height = self.visualization_config["GameWindow"]["min_height"]

        self.buttons_width = self.visualization_config["GameWindow"]["buttons_width"]
        self.buttons_height = self.visualization_config["GameWindow"]["buttons_height"]

        self.order_bar_height = self.visualization_config["GameWindow"][
            "order_bar_height"
        ]

        self.window_width = self.min_width
        self.window_height = self.min_height

        self.main_window = pygame.display.set_mode(
            (self.window_width, self.window_height)
        )

        # self.game_width, self.game_height = 0, 0

        self.images_path = ROOT_DIR / "pygame_gui" / "images"

        self.menu_state = MenuStates.Start
        self.manager: pygame_gui.UIManager

        self.vis = Visualizer(self.visualization_config)
        self.vis.create_player_colors(len(self.player_names))

    def get_window_sizes(self, state: dict):
        counter_positions = np.array([c["pos"] for c in state["counters"]])
        kitchen_width = counter_positions[:, 0].max() + 0.5
        kitchen_height = counter_positions[:, 1].max() + 0.5
        if self.visualization_config["GameWindow"]["WhatIsFixed"] == "window_width":
            game_width = self.visualization_config["GameWindow"]["size"]
            kitchen_aspect_ratio = kitchen_height / kitchen_width
            game_height = int(game_width * kitchen_aspect_ratio)
            grid_size = int(game_width / self.simulator.env.kitchen_width)

        elif self.visualization_config["GameWindow"]["WhatIsFixed"] == "window_height":
            game_height = self.visualization_config["GameWindow"]["size"]
            kitchen_aspect_ratio = kitchen_width / kitchen_height
            game_width = int(game_height * kitchen_aspect_ratio)
            grid_size = int(game_width / self.simulator.env.kitchen_width)

        elif self.visualization_config["GameWindow"]["WhatIsFixed"] == "grid":
            grid_size = self.visualization_config["GameWindow"]["size"]
            game_width, game_height = (
                kitchen_width * grid_size,
                kitchen_height * grid_size,
            )

        else:
            game_width, game_height = 0, 0
            grid_size = 0

        window_width, window_height = (
            game_width + (2 * self.screen_margin),
            game_height + (2 * self.screen_margin),  # bar with orders
        )

        window_width = max(window_width, self.min_width)
        window_height = max(window_height, self.min_height)
        return (
            int(window_width),
            int(window_height),
            int(game_width),
            int(game_height),
            grid_size,
        )

    def handle_keys(self):
        """Handles keyboard inputs. Sends action for the respective players. When a key is held down, every frame
        an action is sent in this function.
        """
        keys = pygame.key.get_pressed()
        for player_idx, key_set in enumerate(self.player_key_sets):
            relevant_keys = [keys[k] for k in key_set.player_keys]
            if any(relevant_keys[:-2]):
                move_vec = np.zeros(2)
                for idx, pressed in enumerate(relevant_keys[:-2]):
                    if pressed:
                        move_vec += key_set.move_vectors[idx]
                if np.linalg.norm(move_vec) != 0:
                    move_vec = move_vec / np.linalg.norm(move_vec)

                action = Action(
                    key_set.name, ActionType.MOVEMENT, move_vec, duration=1 / self.FPS
                )
                self.send_action(action)

    def handle_key_event(self, event):
        """Handles key events for the pickup and interaction keys. Pickup is a single action,
        for interaction keydown and keyup is necessary, because the player has to be able to hold
        the key down.

        Args:
            event: Pygame event for extracting the key action.
        """
        for key_set in self.player_key_sets:
            if event.key == key_set.pickup_key and event.type == pygame.KEYDOWN:
                action = Action(key_set.name, ActionType.PUT, "pickup")
                self.send_action(action)

            if event.key == key_set.interact_key:
                if event.type == pygame.KEYDOWN:
                    action = Action(
                        key_set.name, ActionType.INTERACT, InterActionData.START
                    )
                    self.send_action(action)
                elif event.type == pygame.KEYUP:
                    action = Action(
                        key_set.name, ActionType.INTERACT, InterActionData.STOP
                    )
                    self.send_action(action)

    def init_ui_elements(self):
        self.manager = pygame_gui.UIManager((self.window_width, self.window_height))
        self.manager.get_theme().load_theme(ROOT_DIR / "gui_2d_vis" / "gui_theme.json")

        self.start_button = pygame_gui.elements.UIButton(
            relative_rect=pygame.Rect(
                (
                    (self.window_width // 2) - self.buttons_width // 2,
                    (self.window_height / 2) - self.buttons_height // 2,
                ),
                (self.buttons_width, self.buttons_height),
            ),
            text="Start Game",
            manager=self.manager,
        )
        self.start_button.can_hover()

        self.quit_button = pygame_gui.elements.UIButton(
            relative_rect=pygame.Rect(
                (
                    (self.window_width - self.buttons_width),
                    0,
                ),
                (self.buttons_width, self.buttons_height),
            ),
            text="Quit Game",
            manager=self.manager,
            object_id="#quit_button",
        )
        self.quit_button.can_hover()

        self.reset_button = pygame_gui.elements.UIButton(
            relative_rect=pygame.Rect(
                (
                    (self.screen_margin + self.game_width),
                    self.screen_margin,
                ),
                (self.screen_margin, 100),
            ),
            text="RESET",
            manager=self.manager,
            object_id="#quit_button",
        )
        self.reset_button.can_hover()

        self.finished_button = pygame_gui.elements.UIButton(
            relative_rect=pygame.Rect(
                (
                    (self.window_width - self.buttons_width),
                    (self.window_height - self.buttons_height),
                ),
                (self.buttons_width, self.buttons_height),
            ),
            text="Finish round",
            manager=self.manager,
        )
        self.finished_button.can_hover()

        self.back_button = pygame_gui.elements.UIButton(
            relative_rect=pygame.Rect(
                (
                    (0),
                    (self.window_height - self.buttons_height),
                ),
                (self.buttons_width, self.buttons_height),
            ),
            text="Back to menu",
            manager=self.manager,
        )
        self.back_button.can_hover()

        self.score_label = pygame_gui.elements.UILabel(
            text=f"Score: _",
            relative_rect=pygame.Rect(
                (
                    (0),
                    self.window_height - self.screen_margin,
                ),
                (self.screen_margin * 2, self.screen_margin),
            ),
            manager=self.manager,
            object_id="#score_label",
        )

        layout_file_paths = [
            str(p.name)
            for p in (ROOT_DIR / "game_content" / "layouts").glob("*.layout")
        ]
        assert len(layout_file_paths) != 0, "No layout files."
        dropdown_width, dropdown_height = 200, 40
        self.layout_selection = pygame_gui.elements.UIDropDownMenu(
            relative_rect=pygame.Rect(
                (
                    0,
                    0,
                ),
                (dropdown_width, dropdown_height),
            ),
            manager=self.manager,
            options_list=layout_file_paths,
            starting_option=layout_file_paths[-1],
        )
        self.timer_label = pygame_gui.elements.UILabel(
            text="GAMETIME",
            relative_rect=pygame.Rect(
                (self.screen_margin, self.window_height - self.screen_margin),
                (self.game_width, self.screen_margin),
            ),
            manager=self.manager,
            object_id="#timer_label",
        )

        self.orders_label = pygame_gui.elements.UILabel(
            text="Orders:",
            relative_rect=pygame.Rect(0, 0, self.screen_margin, self.screen_margin),
            manager=self.manager,
            object_id="#orders_label",
        )

        self.conclusion_label = pygame_gui.elements.UILabel(
            text="Your final score was _",
            relative_rect=pygame.Rect(0, 0, self.window_width, self.window_height),
            manager=self.manager,
            object_id="#score_label",
        )

    def draw(self, state):
        """Main visualization function.

        Args:            state: The game state returned by the environment."""
        self.vis.draw_gamescreen(
            self.game_screen,
            state,
            self.game_width,
            self.game_height,
            self.grid_size,
        )

        # self.manager.draw_ui(self.main_window)
        self.update_remaining_time(state["remaining_time"])

        self.vis.draw_orders(
            screen=self.main_window,
            state=state,
            grid_size=self.grid_size,
            width=self.game_width,
            height=self.game_height,
            screen_margin=self.screen_margin,
            config=self.visualization_config,
        )
        self.update_score_label(state)

    def set_window_size(self):
        self.game_screen = pygame.Surface(
            (
                self.game_width,
                self.game_height,
            ),
        )
        self.main_window = pygame.display.set_mode(
            (
                self.window_width,
                self.window_height,
            )
        )

    def reset_window_size(self):
        self.window_width = self.min_width
        self.window_height = self.min_height
        self.game_width = 0
        self.game_height = 0
        self.set_window_size()
        self.init_ui_elements()

    def manage_button_visibility(self):
        match self.menu_state:
            case MenuStates.Start:
                self.back_button.hide()
                self.quit_button.show()
                self.start_button.show()
                self.score_label.hide()
                self.finished_button.hide()
                self.layout_selection.show()
                self.timer_label.hide()
                self.orders_label.hide()
                self.conclusion_label.hide()
            case MenuStates.Game:
                self.start_button.hide()
                self.back_button.hide()
                self.score_label.show()
                self.score_label.show()
                self.finished_button.show()
                self.layout_selection.hide()
                self.timer_label.show()
                self.orders_label.show()
                self.conclusion_label.hide()
            case MenuStates.End:
                self.start_button.hide()
                self.back_button.show()
                self.score_label.hide()
                self.finished_button.hide()
                self.layout_selection.hide()
                self.timer_label.hide()
                self.orders_label.hide()
                self.conclusion_label.show()

    def update_score_label(self, state):
        score = state["score"]
        self.score_label.set_text(f"Score {score}")

    def update_conclusion_label(self, state):
        score = state["score"]
        self.conclusion_label.set_text(f"Your final score is {score}. Hurray!")

    def update_remaining_time(self, remaining_time: float):
        hours, rem = divmod(int(remaining_time), 3600)
        minutes, seconds = divmod(rem, 60)
        display_time = f"{minutes}:{'%02d' % seconds}"
        self.timer_label.set_text(f"Time remaining: {display_time}")

    def setup_environment(self):
        environment_config_path = ROOT_DIR / "game_content" / "environment_config.yaml"
        layout_path = ROOT_DIR / "game_content" / "layouts" / "basic.layout"
        item_info_path = ROOT_DIR / "game_content" / "item_info_debug.yaml"
        with open(item_info_path, "r") as file:
            item_info = file.read()
        with open(layout_path, "r") as file:
            layout = file.read()
        with open(environment_config_path, "r") as file:
            environment_config = file.read()
        creation_json = CreateEnvironmentConfig(
            manager_id=MANAGER_ID,
            number_players=2,
            environment_settings={"all_player_can_pause_game": False},
            item_info_config=item_info,
            environment_config=environment_config,
            layout_config=layout,
        ).model_dump(mode="json")
        # print(CreateEnvironmentConfig.model_validate_json(json_data=creation_json))
        env_info = requests.post(
            f"{self.request_url}/manage/create_env/",
            json=creation_json,
        )
        env_info = env_info.json()
        assert isinstance(env_info, dict), "Env info must be a dictionary"
        self.current_env_id = env_info["env_id"]
        self.player_info = env_info["player_info"]
        for player_id, player_info in env_info["player_info"].items():
            websocket = connect(self.websocket_url + player_info["client_id"])
            websocket.send(
                json.dumps({"type": "ready", "player_hash": player_info["player_hash"]})
            )
            assert json.loads(websocket.recv())["status"] == 200, "not accepted player"
            self.websockets[player_id] = websocket
        self.state_player_id = player_id
        websocket.send(
            json.dumps({"type": "get_state", "player_hash": player_info["player_hash"]})
        )
        state = json.loads(websocket.recv())

        (
            self.window_width,
            self.window_height,
            self.game_width,
            self.game_height,
            self.grid_size,
        ) = self.get_window_sizes(state)

    def start_button_press(self):
        self.menu_state = MenuStates.Game

        self.setup_environment()

        self.set_window_size()

        self.init_ui_elements()
        log.debug("Pressed start button")

        # self.api.set_sim(self.simulator)

    def back_button_press(self):
        self.menu_state = MenuStates.Start
        self.reset_window_size()
        log.debug("Pressed back button")

    def quit_button_press(self):
        self.running = False
        self.menu_state = MenuStates.Start
        log.debug("Pressed quit button")

    def reset_button_press(self):
        requests.post(
            f"{self.request_url}/manage/stop_env",
            json={
                "manager_id": MANAGER_ID,
                "env_id": self.current_env_id,
                "reason": "reset button pressed",
            },
        )

        # self.websocket.send(json.dumps("reset_game"))
        # answer = self.websocket.recv()        log.debug("Pressed reset button")

    def finished_button_press(self):
        requests.post(
            f"{self.request_url}/manage/stop_env",
            json={
                "manager_id": MANAGER_ID,
                "env_id": self.current_env_id,
                "reason": "finish button pressed",
            },
        )
        self.menu_state = MenuStates.End
        self.reset_window_size()
        log.debug("Pressed finished button")

    def send_action(self, action: Action):
        """Sends an action to the game environment.

        Args:
            action: The action to be sent. Contains the player, action type and move direction if action is a movement.
        """
        if isinstance(action.action_data, np.ndarray):
            action.action_data = [
                float(action.action_data[0]),
                float(action.action_data[1]),
            ]
        self.websockets[action.player].send(
            json.dumps(
                {
                    "type": "action",
                    "action": dataclasses.asdict(
                        action, dict_factory=custom_asdict_factory
                    ),
                    "player_hash": self.player_info[action.player]["player_hash"],
                }
            )
        )
        self.websockets[action.player].recv()

    def request_state(self):
        self.websockets[self.state_player_id].send(
            json.dumps(
                {
                    "type": "get_state",
                    "player_hash": self.player_info[self.state_player_id][
                        "player_hash"
                    ],
                }
            )
        )
        # self.websocket.send(json.dumps("get_state"))
        # state_dict = json.loads(self.websocket.recv())
        state = json.loads(self.websockets[self.state_player_id].recv())
        return state

    def disconnect_websockets(self):
        for websocket in self.websockets.values():
            websocket.close()

    def start_pygame(self):
        """Starts pygame and the gui loop. Each frame the game state is visualized and keyboard inputs are read."""
        log.debug(f"Starting pygame gui at {self.FPS} fps")
        pygame.init()
        pygame.font.init()

        pygame.display.set_caption("Simple Overcooked Simulator")

        clock = pygame.time.Clock()

        self.reset_window_size()
        self.init_ui_elements()
        self.manage_button_visibility()

        # Game loop
        self.running = True
        while self.running:
            try:
                time_delta = clock.tick(self.FPS) / 1000.0

                for event in pygame.event.get():
                    if event.type == pygame.QUIT:
                        self.running = False

                        # UI Buttons:
                    if event.type == pygame_gui.UI_BUTTON_PRESSED:
                        match event.ui_element:
                            case self.start_button:
                                self.start_button_press()
                            case self.back_button:
                                self.back_button_press()
                                self.disconnect_websockets()

                            case self.finished_button:
                                self.finished_button_press()
                                self.disconnect_websockets()
                            case self.quit_button:
                                self.quit_button_press()
                                self.disconnect_websockets()
                            case self.reset_button:
                                self.reset_button_press()
                                self.disconnect_websockets()
                                self.start_button_press()

                        self.manage_button_visibility()

                    if (
                        event.type in [pygame.KEYDOWN, pygame.KEYUP]
                        and self.menu_state == MenuStates.Game
                    ):
                        pass
                        self.handle_key_event(event)

                    self.manager.process_events(event)

                    # drawing:
                self.main_window.fill(
                    colors[self.visualization_config["GameWindow"]["background_color"]]
                )
                self.manager.draw_ui(self.main_window)

                match self.menu_state:
                    case MenuStates.Start:
                        pass

                    case MenuStates.Game:
                        state = self.request_state()

                        self.handle_keys()

                        if state["ended"]:
                            self.finished_button_press()
                            self.disconnect_websockets()
                            self.manage_button_visibility()
                        else:
                            self.draw(state)

                            game_screen_rect = self.game_screen.get_rect()
                            game_screen_rect.center = [
                                self.window_width // 2,
                                self.window_height // 2,
                            ]
                            self.main_window.blit(self.game_screen, game_screen_rect)

                    case MenuStates.End:
                        self.update_conclusion_label(state)

                self.manager.update(time_delta)
                pygame.display.flip()

            except (KeyboardInterrupt, SystemExit):
                self.running = False

        pygame.quit()
        sys.exit()


def main(url, port):
    # TODO maybe read the player names and keyboard keys from config file?
    keys1 = [
        pygame.K_LEFT,
        pygame.K_RIGHT,
        pygame.K_UP,
        pygame.K_DOWN,
        pygame.K_SPACE,
        pygame.K_i,
    ]
    keys2 = [pygame.K_a, pygame.K_d, pygame.K_w, pygame.K_s, pygame.K_f, pygame.K_e]

    number_players = 2
    gui = PyGameGUI(
        list(map(str, range(number_players))), [keys1, keys2], url=url, port=port
    )
    gui.start_pygame()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="Overcooked Simulator 2D PyGame Visualization",
        description="PyGameGUI: a PyGame 2D Visualization window.",
        epilog="For further information, see https://scs.pages.ub.uni-bielefeld.de/cocosy/overcooked-simulator/overcooked_simulator.html",
    )

    url_and_port_arguments(parser)
    disable_websocket_logging_arguments(parser)
    args = parser.parse_args()
    setup_logging(enable_websocket_logging=args.enable_websocket_logging)
    main(args.url, args.port)
